package ap.proof.unificationOpt;

import ap.terfor.ConstantTerm

/**
 * Datatype for constants that are bound in a proof tree. Such constants are
 * partially ordered by the order of the binders: a
 * constant c is strictly smaller than another constant d if both constants have
 * been introduced on the same branch, and if c has been introduced earlier 
 * (farther outside) than d.
 *
 * The attribute <code>position</code> tells on which branch the constant is
 * bound (this is necessary, because the same <code>ConstantTerm</code> object
 * can be used and bound multiple times in a proof). If <code>position</code>
 * is an empty list, the constant is not considered bound at all.
 */
case class BoundConstant(const : ConstantTerm, position : List[Int])
           extends PartiallyOrdered[BoundConstant] {
  
  def tryCompareTo[B >: BoundConstant <% PartiallyOrdered[B]](that : B) : Option[Int] =
    that match {
      case that : BoundConstant =>
        if (endsWith(this.position, that.position)) {
          if (this.position.size == that.position.size)
            Some(0)
          else
            Some(1)
        } else if (endsWith(this.position, that.position)) {
          Some(-1)
        } else {
          None
        }
      case _ => None
    }
  
  private def endsWith[A](a : List[A], b : List[A]) : Boolean =
    a.size >= b.size && (a drop (a.size - b.size)) == b
  
}
