package ap.terfor.substitutions;


import ap.util.Debug

object ConstantSubst {
  
  private val AC = Debug.AC_SUBSTITUTIONS
  
  def unapply(s : Substitution) : Option[(ConstantTerm, Term)] = s match {
    case s : ConstantSubst => Some((s.constant, s.replacement))
    case _ => None
  }
  
}

/**
 * Replace a constant with an arbitrary term
 */
class ConstantSubst(private val constant : ConstantTerm,
                    private val replacement : Term,
                    protected [substitutions] val order : TermOrder)
      extends SimpleSubstitution {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(ConstantSubst.AC, order isSortingOf replacement)
  //////////////////////////////////////////////////////////////////////////////     

  protected[substitutions] def passQuantifiers(num : Int) : Substitution =
    new ConstantSubst(constant,
                      // TODO: could be optimised by caching
                      new VariableShiftSubst(0, num, order) (replacement),
                      order)
    
  protected[substitutions] def applyToVariable(v : VariableTerm) : Term = v

  protected[substitutions] def applyToConstant(c : ConstantTerm) : Term =
    if (c == constant) replacement else c

  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean =
    !(t.constants contains constant)

  //////////////////////////////////////////////////////////////////////////////

  def sortBy(newOrder : TermOrder) : ConstantSubst =
    new ConstantSubst(constant, newOrder sort replacement, newOrder)

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String = "[" + constant + " |-> " + replacement + "]"

}
