package ap.terfor.substitutions;

import ap.util.{Debug, Logic}

object VariableShiftSubst {
  
  private val AC = Debug.AC_SUBSTITUTIONS
  
  /**
   * Create functions for shifting the variables in terms or formulas up or
   * down. This is frequently necessary because of De-Brujin indexes 
   */
  def upShifter[A <: TerFor](shift : Int, order : TermOrder) : (A) => A = {
    val shiftUpSubst = new VariableShiftSubst(0, shift, order)
    ((x:A) => shiftUpSubst(x).asInstanceOf[A])
  }
  
  /**
   * Create functions for shifting the variables in terms or formulas up or
   * down. This is frequently necessary because of De-Brujin indexes 
   */
  def downShifter[A <: TerFor](shift : Int, order : TermOrder) = new PartialFunction[A, A] {
    val shiftDownSubst = new VariableShiftSubst(shift, -shift, order)
    def isDefinedAt(x : A) : Boolean =
      Logic.forall(for (VariableTerm(i) <- x.variables.elements)
                   yield i >= shift)
    def apply(x : A) : A = shiftDownSubst(x).asInstanceOf[A]
  }
  
}

/**
 * Shift all variables with an index that is at least <code>offset</code>
 * upwards (<code>shift</code> steps). <code>shift</code> can be negative to
 * realise downwards shifting
 */
class VariableShiftSubst(offset : Int, shift : Int,
                         protected [substitutions] val order : TermOrder)
      extends SimpleSubstitution {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(VariableShiftSubst.AC, offset >= 0 && offset + shift >= 0)
  //////////////////////////////////////////////////////////////////////////////     

  protected[substitutions] def passQuantifiers(num : Int) : Substitution =
    new VariableShiftSubst(offset + num, shift, order)
  
  protected[substitutions] def applyToVariable(v : VariableTerm) : Term = {
    val i = v.index
    if (i < offset) v else VariableTerm(i + shift) 
  }

  protected[substitutions] def applyToConstant(c : ConstantTerm) : Term = c

  protected[substitutions] def isIdentityOn(t : TerFor) : Boolean =
    Logic.forall(for (VariableTerm(i) <- t.variables.elements) yield i < offset)

  //////////////////////////////////////////////////////////////////////////////
    
  def sortBy(newOrder : TermOrder) : VariableShiftSubst =
    new VariableShiftSubst(offset, shift, newOrder)

}
